from http import HTTPStatus
import logging

# Output config
class Config():
    log_level = logging.INFO

    # Base URL for all methods
    API_URL = 'https://api.smartcitizen.me/v0/'
    DEVICES_URL =  API_URL + 'devices/'
    SENSORS_URL = API_URL + 'sensors/'
    MEASUREMENTS_URL = API_URL + 'measurements/'
    FRONTEND_URL = 'https://smartcitizen.me/kits/'
    BASE_POSTPROCESSING_URL='https://raw.githubusercontent.com/fablabbcn/smartcitizen-data/master/'
    API_SEARCH_URL = API_URL + "search?q="

    # Alphasense sensor codes
    _as_sensor_codes =  {
        '132':  'ASA4_CO',
        '133':  'ASA4_H2S',
        '130':  'ASA4_NO',
        '212':  'ASA4_NO2',
        '214':  'ASA4_OX',
        '134':  'ASA4_SO2',
        '162':  'ASB4_CO',
        '133':  'ASB4_H2S',#
        '130':  'ASB4_NO', #
        '202':  'ASB4_NO2',
        '204':  'ASB4_OX',
        '164':  'ASB4_SO2'
    }

    _max_retries = 3
    _rety_interval = 10
    _retry_codes = [
        HTTPStatus.TOO_MANY_REQUESTS,
        HTTPStatus.INTERNAL_SERVER_ERROR,
        HTTPStatus.BAD_GATEWAY,
        HTTPStatus.SERVICE_UNAVAILABLE,
        HTTPStatus.GATEWAY_TIMEOUT,
    ]

    _max_concurrent_requests = 30

config = Config()


from smartcitizen_connector.models import (Device, HardwarePostprocessing, Metric, Postprocessing)
from smartcitizen_connector._config import config
from smartcitizen_connector.tools import logger, safe_get, tf, \
    convert_freq_to_rollup, clean, localise_date, url_checker, process_headers, get_alphasense, \
    get_pt_temp, find_by_field, dict_fmerge
from typing import Optional, List, Dict
from requests import get, post, patch
from aiohttp_retry import RetryClient, ExponentialRetry
from pandas import DataFrame, to_datetime
from datetime import datetime
from os import environ
from pydantic import TypeAdapter
import sys
import json
from math import isnan
from tqdm import trange
from json import dumps, JSONEncoder, loads
import aiohttp
import asyncio
import time

# numpy to json encoder to avoid convertion issues. borrowed from
# https://stackoverflow.com/questions/50916422/python-typeerror-object-of-type-int64-is-not-json-serializable#50916741
class NpEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, integer):
            return int(obj)
        if isinstance(obj, floating):
            return float(obj)
        if isinstance(obj, ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)

def check_blueprint(blueprint_url):
    if blueprint_url is None or not blueprint_url:
        logger.info('No blueprint url')
        return None
    if url_checker(blueprint_url):
        _blueprint = safe_get(blueprint_url).json()
    else:
        logger.info(f'No valid blueprint in url')
        _blueprint = None
    return _blueprint

def check_postprocessing(postprocessing):
    # Postprocessing should be dict or Postprocessing Model
    if postprocessing is None:
        return None, None, False

    if type(postprocessing) == dict:
        _postprocessing = TypeAdapter(Postprocessing).validate_python(postprocessing)
    else:
        _postprocessing = postprocessing

    # Check the url in hardware
    urls = url_checker(_postprocessing.hardware_url)
    # If URL is empty, try prepending base url from config
    if not urls:
        tentative_url = f"{config.BASE_POSTPROCESSING_URL}hardware/{_postprocessing.hardware_url}.json"
    else:
        if len(urls)>1: logger.warning('URLs for postprocessing recipe are more than one, trying first')
        tentative_url = urls[0]

    logger.info(f'Device has postprocessing information:\n{_postprocessing}')

    _ok = True
    # Make hardware postprocessing
    if url_checker(tentative_url):
        try:
            r = safe_get(tentative_url)
        except:
            _ok = False
            _hardware_postprocessing = None
            pass
    else:
        _hardware_postprocessing = None
        _ok = False

    if _ok:
        _hardware_url = tentative_url
        try:
            _hardware_postprocessing = TypeAdapter(HardwarePostprocessing).validate_python(r.json())
        except:
            _ok = False
            _hardware_postprocessing = None
            pass
    else:
        _hardware_url = ''

    return _hardware_url, _hardware_postprocessing, _ok

class SCDevice:

    def __init__(self, id = None, params = None, check_postprocessing=True):
        if id is not None:
            self.id = id
        elif params is not None:
            self.id = params.id
        else:
            raise ValueError("Need at least id or params.id")
        self.params = params
        self.url = f'{config.DEVICES_URL}{self.id}'
        self.page = f'{config.FRONTEND_URL}{self.id}'
        self.method = 'async'
        self.data = DataFrame()
        self._metrics: List[Metric] = []
        self.__load__()
        self.__get_timezone__()
        if check_postprocessing:
            self.__check_postprocessing__()
            self._filled_properties = list()
            self._properties = dict()
            if self.__check_blueprint__():
                if self.__get_metrics__():
                    # TODO Improve how this happens automatically
                    self._filled_properties.append('metrics')
                self.__make_properties__()
        else:
            self._metrics = []
        logger.info(f'Device {self.json.id} initialized')

    def __load__(self):
        r = safe_get(self.url)
        # TODO assess if one day SCDevice can inherit directly from Device
        self.json = TypeAdapter(Device).validate_python(r.json())

    def __get_timezone__(self) -> str:

        if self.json.location.latitude is not None and self.json.location.longitude is not None:
            self.timezone = tf.timezone_at(lng=self.json.location.longitude, lat=self.json.location.latitude)

        logger.info('Device {} timezone is {}'.format(self.id, self.timezone))

        return self.timezone

    def __check_blueprint__(self):
        self._blueprint = check_blueprint(self.blueprint_url)

        return self._blueprint is not None

    def __check_postprocessing__(self):
        logger.info(f'Checking postprocessing of {self.id}')
        if self.json.postprocessing is not None:
            self.json.postprocessing.hardware_url, self._hardware_postprocessing, valid = check_postprocessing(self.json.postprocessing)
        else:
            self._hardware_postprocessing = None
            logger.warning('No postprocessing information')

    def __get_metrics__(self):
        self._metrics = TypeAdapter(List[Metric]).validate_python([y for y in self._blueprint['metrics']])

        # Convert that to metrics now
        if self._hardware_postprocessing is not None:
            for version in self._hardware_postprocessing.versions:
                if version.from_date is not None:
                    if version.from_date > self.last_reading_at:
                        logger.warning('Postprocessing from_date is later than device last_reading_at. Skipping')
                        continue

                for slot in version.ids:
                    metrics = None
                    if slot.startswith('AS'):
                        metric = get_alphasense(slot, version.ids[slot])
                    elif slot.startswith('PT'):
                        metric = get_pt_temp(slot, version.ids[slot])
                    for m in metric:
                        for key, value in m.items():
                            item = find_by_field(self._metrics, key, 'name')
                            if item is None:
                                logger.warning(f'Item not found, {item[0]}')
                                continue
                            item.kwargs = dict_fmerge(item.kwargs, value['kwargs'])
            return True

    def __make_properties__(self):
        for item, value in self._blueprint.items():
            if item in self._filled_properties:
                self._properties[item] = self.__getattribute__(item)
            else:
                self._properties[item] = value

    async def get_datum(self, semaphore, session, url, headers, sensor_id, resample, frequency, rename)->Dict:
        async with semaphore:
            retry_client = RetryClient(client_session=session)
            async with retry_client.get(url, headers = headers) as resp:

                datum = await resp.json()
                sensor_name = find_by_field(self.json.data.sensors, sensor_id, 'id').name

                if 'readings' not in datum:
                    logger.warning(f"Device: {self.json.id}- No readings in request for sensor: {sensor_id}: {sensor_name}")
                    logger.warning(f"Response code: {resp}")
                    return None

                if datum['readings'] == []:
                    logger.warning(f"Device: {self.json.id} - No data in request for sensor: {sensor_id}: {sensor_name}")
                    return None

                logger.info(f"Device: {self.json.id} - Got readings for sensor: {sensor_id}: {sensor_name}")

                # Make a Dataframe
                # Set index
                df_sensor = DataFrame(datum['readings']).set_index(0)
                # Set columns
                if rename:
                    df_sensor.columns = [sensor_name]
                else:
                    df_sensor.columns = [str(sensor_id)]
                # Localise index
                df_sensor.index = localise_date(df_sensor.index, self.timezone)
                # Sort it just in case
                df_sensor.sort_index(inplace=True)
                # Remove duplicates
                df_sensor = df_sensor[~df_sensor.index.duplicated(keep='first')]
                # Drop unnecessary columns
                df_sensor.drop([i for i in df_sensor.columns if 'Unnamed' in i], axis=1, inplace=True)
                # Check for weird things in the data
                df_sensor = df_sensor.astype(float, errors='ignore')
                # Resample
                if (resample):
                    df_sensor = df_sensor.resample(frequency).mean()

                return df_sensor

    async def get_data(self,
        min_date: Optional[datetime] = None,
        max_date: Optional[datetime] = None,
        frequency: Optional[str] = '1Min',
        clean_na: Optional[str] = None,
        resample: Optional[bool] = False,
        rename: Optional[bool] = True)->DataFrame:

        if 'SC_ADMIN_BEARER' in environ:
            logger.info('Admin Bearer found, using it')
            headers = {'Authorization':'Bearer ' + environ['SC_ADMIN_BEARER']}
        elif 'SC_BEARER' in environ:
            logger.info('Bearer found in environment, using it.')
            # TODO make this explicit
            headers = {'Authorization':'Bearer ' + environ['SC_BEARER']}
        else:
            logger.warning('No Bearer not found, you might get throttled!')
            headers = None

        logger.info(f'Make sure we are up to date')
        self.__load__()
        logger.info(f'Requesting data from SC API')
        logger.info(f'Device ID: {self.id}')
        rollup = convert_freq_to_rollup(frequency)
        logger.info(f'Using rollup: {rollup}')

        if self.timezone is None:
            logger.warning('Device does not have timezone set, skipping')
            return self.data

        # Check start date and end date
        # Converting to UTC by passing None
        # https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.dt.tz_convert.html
        if min_date is not None:
            min_date = localise_date(to_datetime(min_date), 'UTC')
            logger.info(f'Min Date: {min_date}')
        else:
            min_date = localise_date(to_datetime('2001-01-01'), 'UTC')
            logger.info("No min_date specified")

        if max_date is not None:
            max_date = localise_date(to_datetime(max_date), 'UTC')
            logger.info(f'Max Date: {max_date}')

        # Trim based on actual data available
        if min_date is not None and self.json.last_reading_at is not None:
            if min_date > self.json.last_reading_at:
                logger.warning(f'Device {self.id} request would yield empty data (min_date). Returning')
                return self.data

        if max_date is not None and self.json.created_at is not None:
            if max_date < self.json.created_at:
                logger.warning(f'Device {self.id} request would yield empty data (max_date). Returning')
                return self.data

        if max_date is not None and self.json.last_reading_at is not None:
            if max_date > self.json.last_reading_at:
                logger.warning('Trimming max_date to last reading')
                max_date = self.json.last_reading_at

        logger.info(f'Device {self.id} timezone: {self.timezone}')

        if not self.json.data.sensors:
            logger.info(f'Device {self.id} is empty')
            return self.data
        else: logger.info(f"Sensor IDs: {[f'{sensor.name}: {sensor.id}' for sensor in self.json.data.sensors]}")

        df = DataFrame()
        logger.info(f'Requesting device {self.id} from {min_date} to {max_date}')

        semaphore = asyncio.Semaphore(config._max_concurrent_requests)
        async with aiohttp.ClientSession() as session:

            tasks = []
            for sensor in self.json.data.sensors:

                # Request sensor per ID
                url = self.url + '/readings?'

                if min_date is not None: url += f'from={min_date}'
                if max_date is not None: url += f'&to={max_date}'

                url += f'&rollup={rollup}'
                url += f'&sensor_id={sensor.id}'
                url += '&function=avg'

                tasks.append(asyncio.ensure_future(self.get_datum(semaphore, session, url, headers, sensor.id, resample, frequency, rename)))

            dfs_sensor = await asyncio.gather(*tasks)

        # Process received data
        for df_sensor in dfs_sensor:
            if df_sensor is None: continue
            # Combine in the main df
            df = df.combine_first(df_sensor)

        try:
            df = df.reindex(df.index.rename('TIME'))
            df = clean(df, clean_na, how = 'all')
            self.data = df
        except:
            logger.error(f'Problem closing up the API dataframe for {self.id}')
            pass

        logger.info(f'Device {self.id} loaded successfully from API')
        return self.data

    async def post_data(self, columns = 'sensors', rename = None, clean_na = 'drop', chunk_size = 500, dry_run = False, max_retries = 2, delay = None):
        '''
            POST self.data in the SmartCitizen API
            Parameters
            ----------
                columns: List or string
                    'sensors'
                    If string, either 'sensors' or 'metrics. Empty string is 'sensors' + 'metrics'
                    If list, list containing column names.
                clean_na: string, optional
                    'drop'
                    'drop', 'fill'
                chunk_size: integer
                    chunk size to split resulting pandas DataFrame for posting readings
                dry_run: boolean
                    False
                    Post the payload to the API or just return it
                max_retries: int
                    2
                    Maximum number of retries per chunk
            Returns
            -------
                True if the data was posted succesfully
        '''

        if self.data is None:
            logger.error('No data to post, ignoring')
            return False

        if 'SC_BEARER' not in environ:
            logger.error('Cannot post without Auth Bearer')
            return False

        if 'SC_ADMIN_BEARER' in environ:
            logger.info('Using admin Bearer')
            bearer = environ['SC_ADMIN_BEARER']
        else:
            bearer = environ['SC_BEARER']

        headers = {'Authorization':'Bearer ' + bearer, 'Content-type': 'application/json'}
        post_ok = True

        if columns == 'sensors':
            _columns = self.json.data.sensors
            # TODO - when a device has been processed, data will be there for metrics
        elif columns == 'metrics':
            _columns = self._metrics
        elif type(columns) is list:
            _columns = list()
            for column in columns:
                item = find_by_field(self.json.data.sensors + self._metrics, column, 'name')
                if item is not None:
                    _columns.append(item)
        else:
            _columns = self.json.data.sensors + self._metrics

        if rename is None:
            logger.info('Renaming not required')
            _rename = dict()
            for column in _columns:
                _rename[column.name] = column.name
        else:
            _rename = rename

        async with aiohttp.ClientSession() as session:

            tasks = []
            for column in _columns:
                if _rename[column.name] not in self.data:
                    logger.warning(f'{_rename[column.name]} not in data')
                    continue
                if column.id is None:
                    logger.warning(f'{column.name} has no id')
                    continue
                # Get only post data
                df = DataFrame(self.data[_rename[column.name]]).copy()
                # Rename to ID to be able to post
                logger.info(f'Adding {_rename[column.name]} ({column.id}) to post list')
                df.rename(columns={_rename[column.name]: column.id}, inplace = True)
                url = f'{self.url}/readings'
                # Append task
                tasks.append(asyncio.ensure_future(self.post_datum(session, headers, url, df,
                    clean_na = clean_na, chunk_size = chunk_size, dry_run = dry_run,
                    max_retries = max_retries, delay=delay)))

            posts_ok = await asyncio.gather(*tasks)

        return not(False in posts_ok)

    async def post_datum(self, session, headers, url, df, clean_na = 'drop', chunk_size = 500, dry_run = False, max_retries = 2, delay = None):
        '''
            POST external pandas.DataFrame to the SmartCitizen API
            Parameters
            ----------
                session: aiohttp.ClientSession
                headers: dict
                    Auth headers
                df: pandas DataFrame
                    Contains data in a DataFrame format.
                    Data is posted using the column names of the dataframe
                    Data is posted in UTC TZ so dataframe needs to have located
                    timestamp
                clean_na: string, optional
                    'drop'
                    'drop', 'fill'
                chunk_size: integer
                    chunk size to split resulting pandas DataFrame for posting readings
                dry_run: boolean
                    False
                    Post the payload to the API or just return it
                max_retries: int
                    2
                    Maximum number of retries per chunk
            Returns
            -------
                True if the data was posted succesfully
        '''
        # Clean df of nans
        df = clean(df, clean_na, how = 'all')
        logger.info(f'Posting to {url}')
        logger.info(f'Sensor ID: {list(df.columns)[0]}')
        df.index.name = 'recorded_at'

        # Split the dataframe in chunks
        chunked_dfs = [df[i:i+chunk_size] for i in range(0, df.shape[0], chunk_size)]
        if len(chunked_dfs) > 1: logger.info(f'Splitting post in chunks of size {chunk_size}')

        for i in trange(len(chunked_dfs), file=sys.stdout,
                        desc=f"Posting data for {self.id}..."):

            chunk = chunked_dfs[i].copy()

            # Prepare json post
            payload = {"data":[]}
            for item in chunk.index:
                payload["data"].append(
                    {
                        "recorded_at": localise_date(item, 'UTC').strftime('%Y-%m-%dT%H:%M:%SZ'),
                        "sensors": [{
                            "id": column,
                            "value": chunk.loc[item, column]
                        } for column in chunk.columns if not isnan(chunk.loc[item, column])]
                    }
                )

            if dry_run:
                logger.info(f'Dry run request to: {self.url}/readings for chunk ({i+1}/{len(chunked_dfs)})')
                jsd = dumps(payload, indent = 2, cls = NpEncoder)
                logger.debug(jsd)
                return jsd

            post_ok = False
            retries = 0

            while post_ok == False and retries < max_retries:
                if delay is not None: time.sleep(delay)
                response = post(url, data = dumps(payload, cls = NpEncoder), headers = headers)

                if response.status_code == 200 or response.status_code == 201:
                    post_ok = True
                    break
                else:
                    retries += 1
                    logger.warning (f'Chunk ({i+1}/{len(chunked_dfs)}) post failed. \
                            API responded {response.status_code}.\
                            Retrying ({retries}/{max_retries}')

            if (not post_ok) or (retries == max_retries):
                logger.error (f'Chunk ({i+1}/{len(chunked_dfs)}) post failed. \
                        API responded {response.status_code}.\
                        Reached max_retries')
                return False

        return True

    def patch_postprocessing(self, dry_run = False):
        '''
            PATCH postprocessing info into the device in the SmartCitizen API
            Updates all the post info. Changes need to be made info the keys of the postprocessing outside of here

            # Example postprocessing:
            # {
            #   "blueprint_url": "https://github.com/fablabbcn/smartcitizen-data/blob/master/blueprints/sc_21_station_module.json",
            #   "hardware_url": "https://raw.githubusercontent.com/fablabbcn/smartcitizen-data/master/hardware/SCAS210001.json",
            #   "latest_postprocessing": "2020-10-29T08:35:23Z"
            # }
        '''

        if 'SC_ADMIN_BEARER' not in environ:
            logger.error('Cannot post without Admin Auth Bearer')
            return

        headers = {'Authorization':'Bearer ' + environ['SC_ADMIN_BEARER'],
                   'Content-type': 'application/json'}

        post = {"postprocessing_attributes": loads(self.json.postprocessing.model_dump_json())}

        if dry_run:
            logger.info(f'Dry run request to: {self.url}/')
            return dumps(post)

        logger.info(f'Posting postprocessing_attributes:\n {post} to {self.url}')
        response = patch(f'{self.url}/',
                         data = dumps(post), headers = headers)

        if response.status_code == 200 or response.status_code == 201:
            logger.info(f"Postprocessing posted")
            return True
        else:
            logger.info(f"API responded with {response.status_code}")

        return False

    @property
    def blueprint_url(self):
        if self.json.postprocessing is None:
            return None
        if url_checker(self.json.postprocessing.blueprint_url):
            return self.json.postprocessing.blueprint_url
        elif url_checker(self.json.postprocessing.hardware_url):
            return self._hardware_postprocessing.blueprint_url
        else:
            return None

    @property
    def blueprint(self):
        return self._blueprint

    @property
    def properties(self):
        return self._properties

    @property
    def metrics(self):
        return [metric.model_dump() for metric in self._metrics]

    @property
    def sensors(self):
        return [sensor.model_dump() for sensor in self.json.data.sensors]

    @property
    def hardware_postprocessing(self):
        if self._hardware_postprocessing is None:
            return None
        return self._hardware_postprocessing.model_dump()

    @property
    def postprocessing(self):
        if self.json.postprocessing is None:
            return None
        return self.json.postprocessing.model_dump()

    @property
    def latest_postprocessing(self):
        if self.json.postprocessing is not None:
            return self.json.postprocessing.latest_postprocessing
        else:
            return None

    def update_latest_postprocessing(self, date):
        if self.json.postprocessing is not None:
            # Add latest postprocessing rounded up with
            # frequency so that we don't end up in
            # and endless loop processing only the latest data line
            # (minute vs. second precission of the data)
            try:
                self.json.postprocessing.latest_postprocessing = date.to_pydatetime()
            except:
                return False
            else:
                logger.info(f"Updated latest_postprocessing to: {self.latest_postprocessing}")
                return True
        logger.info('Nothing to update')
        return True

    @property
    def last_reading_at(self):
        return self.json.last_reading_at

def get_devices():
    isn = True
    result = list()
    url = config.DEVICES_URL
    while isn:
        r = get(url)
        r.raise_for_status()
        # If status code OK, retrieve data
        h = process_headers(r.headers)
        result += TypeAdapter(List[Device]).validate_python(r.json())

        if 'next' in h:
            if h['next'] == url: isn = False
            elif h['next'] != url: url = h['next']
        else:
            isn = False
    return result



from smartcitizen_connector._config import config
from smartcitizen_connector.tools import *
from typing import Optional, List, Dict
from pandas import DataFrame
from os import environ
from requests import get

def global_search(value: Optional[str] = None) -> DataFrame:
    """
    Gets devices from Smart Citizen API based on basic search query values,
    searching both Users and Devices at the same time.
    Global search documentation: https://developer.smartcitizen.me/#global-search
    Parameters
    ----------
        value: string
            None
            Query to fit
            For null, not_null values, use 'null' or 'not_null'
    Returns
    -------
        A list of kit IDs that comply with the requirements, or the full df, depending on full.
    """
    if 'SC_ADMIN_BEARER' in environ:
        logger.info('Admin Bearer found, using it')
        headers = {'Authorization':'Bearer ' + environ['SC_ADMIN_BEARER']}
    elif 'SC_BEARER' in environ:
        logger.info('Bearer found in environment, using it.')
        # TODO make this explicit
        headers = {'Authorization':'Bearer ' + environ['SC_BEARER']}
    else:
        logger.warning('No Bearer not found, you might get throttled!')
        headers = None

    # Value check
    if value is None: logger.error(f'Value needs a value, {value} supplied'); return None

    url = config.API_SEARCH_URL  + f'{value}'

    df = DataFrame()
    isn = True
    while isn:
        r = get(url, headers = headers)
        r.raise_for_status()
        # If status code OK, retrieve data
        h = process_headers(r.headers)
        df = df.combine_first(DataFrame(r.json()).set_index('id'))

        if 'next' in h:
            if h['next'] == url: isn = False
            elif h['next'] != url: url = h['next']
        else:
            isn = False

    return df

def search_by_query(endpoint: Optional[str] = 'devices',
    search_items: List[Dict] = None) -> DataFrame:
        # key: Optional[str] = '',
        # search_matcher: Optional[str] = '',
        # value: Optional[str] = None]) -> DataFrame:
    """
    Gets devices from Smart Citizen API based on ransack parameters
    Basic query documentation: https://developer.smartcitizen.me/#basic-searching
    Parameters
    ----------
        endpoint: string
            'devices'
            Endpoint to perform the query at (see docs)
        search_items: List[Dict]
            Required keys for each dictionary below:
            key: string
                ''
                Query key according to the basic query documentation.
            search_matcher: string
                ''
                Ransack search_matcher:
                https://activerecord-hackery.github.io/ransack/getting-started/search-matches/
            value: string
                None
                Query to fit
                For null, not_null values, use 'null' or 'not_null'. In this case ignores search_matcher
    Returns
    -------
        DataFrame with devices
    """

    if 'SC_ADMIN_BEARER' in environ:
        logger.info('Admin Bearer found, using it')
        headers = {'Authorization':'Bearer ' + environ['SC_ADMIN_BEARER']}
    elif 'SC_BEARER' in environ:
        logger.info('Bearer found in environment, using it.')
        # TODO make this explicit
        headers = {'Authorization':'Bearer ' + environ['SC_BEARER']}
    else:
        logger.warning('No Bearer not found, you might get throttled!')
        headers = None

    url = f'{config.API_URL}{endpoint}/?'
    url_queries = 0
    for search_item in search_items:
        # Value check
        key = search_item['key']
        value = search_item['value']

        if value is None:
            logger.error(f'Value needs a value, {value} supplied')
            return None

        if url_queries:
            url += '&'

        if value == 'null' or value == 'not_null':
            url += f'q[{key}_{value}]=1'
        else:
            search_matcher = search_item['search_matcher']
            url += f'q[{key}_{search_matcher}]={value}'

        url_queries += 1


    df = DataFrame()
    isn = True
    logger.info(f'Getting: {url}')
    while isn:
        r = get(url, headers = headers)
        r.raise_for_status()
        # If status code OK, retrieve data
        h = process_headers(r.headers)
        if r.json() == []: return None
        df = df.combine_first(DataFrame(r.json()).set_index('id'))

        if 'next' in h:
            if h['next'] == url: isn = False
            elif h['next'] != url: url = h['next']
        else:
            isn = False

    return df


from datetime import datetime
from typing import Optional, List, Dict
from pydantic import BaseModel

class Measurement(BaseModel):
    id: int
    name: str
    description: str

class Metric(BaseModel):
    id: Optional[int] = None
    name: str
    description: Optional[str] = ''
    module: Optional[str] = "scdata.device.process"
    function: str
    unit: Optional[str] = ''
    post: Optional[bool] = False
    args: Optional[dict] = None
    kwargs: Optional[dict] = None

class Sensor(BaseModel):
    id: int
    name: str
    description: str
    unit: Optional[str] = None
    measurement_id: Optional[int] = None
    measurement: Optional[Measurement] = None
    value: Optional[float] = None
    prev_value: Optional[float] = None
    last_reading_at: Optional[datetime] = None
    tags: Optional[List[str]] = []
    default_key: Optional[str] = []

class Kit(BaseModel):
    id: int
    slug: str
    name: str
    description: str
    created_at: datetime
    updated_at: datetime
    sensors: Optional[List[Sensor]] = None

class Owner(BaseModel):
    id: int
    username: str
    role: Optional[str] = ""
    devices: Optional[List[str]] = None

class Location(BaseModel):
    city: str
    country_code: str
    country: str
    exposure: Optional[str] = None
    elevation: Optional[float] = None
    geohash: Optional[str] = None
    latitude: Optional[float] = None
    longitude: Optional[float] = None

class HardwareInfo(BaseModel):
    id: str
    mac: str
    time: str
    esp_bd: str
    hw_ver: str
    sam_bd: str
    esp_ver: str
    sam_ver: str
    rcause: Optional[str] = None

class HardwareVersion(BaseModel):
    from_date: Optional[datetime] = None
    to_date: Optional[datetime] = None
    ids: Optional[dict] = None

class HardwarePostprocessing(BaseModel):
    blueprint_url: Optional[str] = None
    description: Optional[str] = None
    versions: Optional[List[HardwareVersion]] = None
    forwarding: Optional[str] = None

class Postprocessing(BaseModel):
    id: Optional[int] = None
    blueprint_url: Optional[str] = None
    hardware_url: Optional[str] = None
    forwarding_params: Optional[int] = None
    meta: Optional[str] = None
    latest_postprocessing: Optional[datetime] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

class Data(BaseModel):
    sensors: Optional[List[Sensor]] = None

class Notifications(BaseModel):
    low_battery: bool
    stopped_publishing: bool

class Device(BaseModel):
    id: int
    uuid: str
    name: str
    description: str
    state: str
    postprocessing: Optional[Postprocessing] = None
    hardware_info: Optional[HardwareInfo] = None
    system_tags: List[str]
    user_tags: List[str]
    is_private: bool
    notify: Notifications
    last_reading_at: Optional[datetime] = None
    created_at: Optional[datetime] = None
    updated_at: datetime
    owner: Owner
    data: Data
    location: Optional[Location]= None
    kit: Optional[Kit] = None
    device_token: Optional[str] = ''



from pandas import to_datetime
from timezonefinder import TimezoneFinder
from datetime import datetime
from smartcitizen_connector._config import config
from typing import Optional
from termcolor import colored
from requests.exceptions import HTTPError
from requests import get
import re
import logging
import sys
import time

tf = TimezoneFinder()

freq_2_rollup_lut = (
    ['A', 'y'],
    ['M', 'M'],
    ['W', 'w'],
    ['D', 'd'],
    ['H', 'h'],
    ['Min', 'm'],
    ['S', 's'],
    ['ms', 'ms']
)

def clean(df, clean_na = None, how = 'all'):
    """
    Helper function for cleaning nan in a pandas.   Parameters
    ----------
        df: pandas.DataFrame to clean
        clean_na: None or string
            type of nan cleaning. If not None, can be 'drop' or 'fill'
        how: 'string'
            Same as how in dropna, fillna. Can be 'any', or 'all'
    Returns
    -------
        Clean dataframe
    """

    if clean_na is not None:
        if clean_na == 'drop':
            df.dropna(axis = 0, how = how, inplace = True)
        elif clean_na == 'fill':
            df = df.fillna(method = 'bfill').fillna(method = 'ffill')
    return df

def convert_freq_to_rollup(freq):
    """
    Helper function for converting a pandas freq into a rollup of SC API's
    ----------
        freq: str freq from pandas
    Returns
    -------
        rollup: str rollup from SC
    """
    rollup_unit = None
    # Convert freq from pandas to SC API's
    for index, letter in enumerate(freq):
        try:
            aux = int(letter)
        except:
            index_first = index
            letter_first = letter
            rollup_value = freq[:index_first]
            freq_unit = freq[index_first:]
            break

    for item in freq_2_rollup_lut:
        if item[0] == freq_unit:
            rollup_unit = item[1]
            break

    if rollup_unit is None:
        return None
    rollup = rollup_value + rollup_unit
    return rollup

def localise_date(date, timezone, tzaware=True):
    """
    Localises a date if it's tzinfo is None, otherwise converts it to it.
    If the timestamp is tz-aware, converts it as well
    Parameters
    ----------
        date: string or datetime
            Date
        timezone: string
            Timezone string. i.e.: 'Europe/Madrid'
    Returns
    -------
        The date converted to 'UTC' and localised based on the timezone
    """
    if date is not None:
        # Per default, we consider that timestamps are tz-aware or UTC.
        # If not, preprocessing should be done to get there
        result_date = to_datetime(date, utc = tzaware)
        if result_date.tzinfo is not None:
            result_date = result_date.tz_convert(timezone)
        else:
            result_date = result_date.tz_localize(timezone)
    else:
        result_date = None

    return result_date

class CutsomLoggingFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_min = "[%(asctime)s] - %(name)s - %(levelname)s - %(message)s"
    format_deb = "[%(asctime)s] - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: grey + format_min + reset,
        logging.INFO: grey + format_min + reset,
        logging.WARNING: yellow + format_min + reset,
        logging.ERROR: red + format_deb + reset,
        logging.CRITICAL: bold_red + format_deb + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)

logger = logging.getLogger('smartcitizen_connector')
logger.setLevel(config.log_level)
ch = logging.StreamHandler(sys.stdout)
ch.setLevel(config.log_level)
ch.setFormatter(CutsomLoggingFormatter())
logger.addHandler(ch)

def set_logger_level(level=logging.DEBUG):
    logger.setLevel(level)

''' Directly from
https://www.geeksforgeeks.org/python-check-url-string/
'''
def url_checker(string):
    if string is not None:
        regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
        url = re.findall(regex,string)
        return [x[0] for x in url]
    else:
        return []

def process_headers(headers):
    result = {}
    if 'total' in headers: result['total_pages'] = headers['total']
    if 'per-page' in headers: result['per_page'] = headers['per-page']
    if 'link' in headers:
        for item in headers.get('link').split(','):
            chunk = item.replace(' ', '').split(';')
            if 'rel' in chunk[1]:
                which = chunk[1].replace('"', '').split('=')[1]
                if which == 'next':
                    result['next'] = chunk[0].strip('<').strip('>')
                elif which == 'last':
                    result['last'] = chunk[0].strip('<').strip('>')
                elif which == 'prev':
                    result['prev'] = chunk[0].strip('<').strip('>')
                elif which == 'first':
                    result['first'] = chunk[0].strip('<').strip('>')
    return result

def safe_get(url):
    for n in range(config._max_retries):
        try:
            r = get(url)
            r.raise_for_status()
        except HTTPError as exc:
            code = exc.response.status_code

            if code in config._retry_codes:
                time.sleep(config._retry_interval)
                continue

            raise
    return r

def get_alphasense(slot, sensor_id):
    result = list()

    # Alphasense type - AAN 803-04
    as_type = config._as_sensor_codes[sensor_id[0:3]]
    channel = as_type[as_type.index('_')+1:]
    metric = channel
    if channel == 'OX':
        metric = 'O3'

    # Get working and auxiliary electrode names
    wen = f"ADC_{slot.strip('AS_')[:slot.index('_')]}_{slot.strip('AS_')[slot.index('_')+1]}"
    aen = f"ADC_{slot.strip('AS_')[:slot.index('_')]}_{slot.strip('AS_')[slot.index('_')+2]}"

    # Simply fill it up
    logger.info(f'{metric} found in blueprint metrics, filling up with hardware info')

    result.append({metric: {
        'kwargs': {
            'we': wen,
            'ae': aen,
            't': 'EC_SENSOR_TEMP',
            'alphasense_id': str(sensor_id)}}})

    # Add channel name for traceability
    result.append({f'{channel}_WE': {'kwargs': {'channel': wen}}})
    result.append({f'{channel}_AE': {'kwargs': {'channel': aen}}})

    return result

def get_pt_temp(slot, sensor_id):
    result = list()

    # Get working and auxiliary electrode names
    pt1000plus = f"ADC_{slot.strip('PT_')[:slot.index('_')]}_{slot.strip('PT_')[slot.index('_')+1]}"
    pt1000minus = f"ADC_{slot.strip('PT_')[:slot.index('_')]}_{slot.strip('PT_')[slot.index('_')+2]}"

    metric = 'ASPT1000'

    # Simply fill it up
    logger.info(f'{metric} found in blueprint metrics, filling up with hardware info')

    result.append({metric: {
        'kwargs': {
            'pt1000plus': pt1000plus,
            'pt1000minus': pt1000minus,
            'afe_id': str(sensor_id)}}})

    # Add channel name for traceability
    result.append({f'PT1000_POS': {'kwargs': {'channel': pt1000plus}}})

    return result

def find_by_field(models, value, field):
    try:
        item = next(model for _, model in enumerate(models) if model.__getattribute__(field) == value)
    except StopIteration:
        logger.exception(f'Column {field} not in models')
        pass
    else:
        return item
    return None

def dict_fmerge(base_dct, merge_dct, add_keys=True):
    """
    Recursive dict merge.
    From: https://gist.github.com/CMeza99/5eae3af0776bef32f945f34428669437
    Parameters
    ----------
        base_dct: dict
            Dict onto which the merge is executed
        merge_dct: dict
            Dict merged into base_dct
        add_keys: bool
            True
            Whether to add new keys
    Returns
    -------
        Updated dict
    """
    rtn_dct = base_dct.copy()
    if add_keys is False:
        merge_dct = {key: merge_dct[key] for key in set(rtn_dct).intersection(set(merge_dct))}

    rtn_dct.update({
        key: dict_fmerge(rtn_dct[key], merge_dct[key], add_keys=add_keys)
        if isinstance(rtn_dct.get(key), dict) and isinstance(merge_dct[key], dict)
        else merge_dct[key]
        for key in merge_dct.keys()
    })

    return rtn_dct

def dict_unpack(row, column, key):
    try:
        return row[column][key]
    except:
        pass

    return ''


from smartcitizen_connector.models import (Sensor, Measurement)
from smartcitizen_connector._config import config
from smartcitizen_connector.tools import *
from pydantic import TypeAdapter
from typing import List
from requests import get

class SCSensor:
    id: int
    url: str
    page: str
    json: Sensor

    def __init__(self, id):
        self.id = id
        self.url = f'{config.SENSORS_URL}{self.id}'
        self.method = 'async'
        r = self.__safe_get__(self.url)
        self.json = TypeAdapter(Sensor).validate_python(r.json())

    def __safe_get__(self, url):
        r = get(url)
        r.raise_for_status()

        return r

def get_sensors():
    isn = True
    result = list()
    url = config.SENSORS_URL
    while isn:
        r = get(url)
        r.raise_for_status()
        # If status code OK, retrieve data
        h = process_headers(r.headers)
        result += TypeAdapter(List[Sensor]).validate_python(r.json())

        if 'next' in h:
            if h['next'] == url: isn = False
            elif h['next'] != url: url = h['next']
        else:
            isn = False
    return result



"""Python package building configuration."""

from glob import glob
from os.path import basename, splitext

from setuptools import find_packages, setup
import sys

if sys.version_info < (3,0):
    sys.exit("This package requires python 3.")

PROJECT_URLS = {
    "Documentation": "https://docs.smartcitizen.me/",
    "Source Code": "https://github.com/fablabbcn/smartcitizen-connector",
}

setup(
    name="smartcitizen-connector",
    version="1.0.6",
    description="Python connector to download information collected in SmartCitizen API",
    author="oscgonfer",
    license="GNU General Public License v3",
    keywords=['sensors', 'Smart Citizen'],
    long_description = open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/fablabbcn/smartcitizen-connector",
    project_urls=PROJECT_URLS,
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Utilities",
        "Natural Language :: English",
    ],
    install_requires=["pydantic",
        "requests",
        "pandas",
        "timezonefinder",
        "urllib3",
        "aiohttp",
        "aiohttp-retry",
        "termcolor",
        "tqdm"],
    setup_requires=['wheel'],
    zip_safe=False
)



from .models import (Sensor, Measurement, Kit, Owner, Location,
                     HardwareInfo, Postprocessing, Data, Device)
from .device import SCDevice#, get_devices
from .sensor import SCSensor, get_sensors
from .measurement import SCMeasurement, get_measurements
from .search import search_by_query, global_search

__all__ = [
    "Device",
    "Kit",
    "Sensor",
    "Measurement",
    "Owner",
    "Location",
    "Data",
    "Postprocessing",
    "HardwareInfo"
    ]

__version__ = '1.0.6'
